/* ---------------------------------------------------------------------------
    Copyright (c) 2003-2013 Micrel, Inc.
    Copyright (c) 2015 Microchip Technology Inc.
   ---------------------------------------------------------------------------

    NdisDevice.h - NDIS driver device definitions.

    Author      Date        Description
    THa         01/26/11    Ported from NDIS 5.0 driver.
    THa         05/02/11    Added burst lengths.
    THa         05/11/11    Added UDP test limit.
   ---------------------------------------------------------------------------
*/


#ifndef __NDIS_DEVICE_H
#define __NDIS_DEVICE_H


/* -------------------------------------------------------------------------- */

#ifdef DEF_KS8842
#define NDIS_NAME  "KSZ8842"
#else
#define NDIS_NAME  "KSZ8841"
#endif

#define MEM_TAG  'rciM'

/* -------------------------------------------------------------------------- */

#if 0
#define NDISVER  60
#endif

#define DRIVER_MAJOR_VERSION  1
#define DRIVER_MINOR_VERSION  22

/* -------------------------------------------------------------------------- */

//
// Message verbosity: lower values indicate higher urgency
//
#define MP_LOUD    4
#define MP_INFO    3
#define MP_TRACE   2
#define MP_WARNING 1
#define MP_ERROR   0

extern INT MPDebugLevel;

#if DBG
#define DEBUGP(Level, Fmt) \
{ \
    if (Level <= MPDebugLevel) \
    { \
        DbgPrint Fmt; \
    } \
}
#else
#define DEBUGP(Level, Fmt)
#endif

MINIPORT_INITIALIZE                 MiniportInitializeEx;
MINIPORT_HALT                       MiniportHaltEx;
MINIPORT_UNLOAD                     MiniportDriverUnload;
MINIPORT_DEVICE_PNP_EVENT_NOTIFY    MiniportDevicePnPEventNotify;

MINIPORT_INTERRUPT_DPC              MiniportHandleInterrupt;
MINIPORT_ISR                        MiniportISR;
MINIPORT_DISABLE_INTERRUPT          MiniportDisableInterruptEx;
MINIPORT_ENABLE_INTERRUPT           MiniportEnableInterruptEx;

MINIPORT_RETURN_NET_BUFFER_LISTS    MiniportReturnNetBufferLists;

MINIPORT_CHECK_FOR_HANG             MiniportCheckForHangEx;
MINIPORT_PAUSE                      MiniportPause;
MINIPORT_RESTART                    MiniportRestart;
MINIPORT_RESET                      MiniportResetEx;
MINIPORT_SHUTDOWN                   MiniportShutdownEx;

MINIPORT_OID_REQUEST                MiniportOidRequest;
MINIPORT_CANCEL_OID_REQUEST         MiniportCancelOidRequest;

MINIPORT_SEND_NET_BUFFER_LISTS      MiniportSendNetBufferLists;
MINIPORT_CANCEL_SEND                MiniportCancelSend;
MINIPORT_PROCESS_SG_LIST            MiniportProcessSGList;

/* -------------------------------------------------------------------------- */

#include "hardware.h"

#ifdef DBG
#if 1
#define DEBUG_MINIPORT_HALT
#endif
#endif

#include <pshpack1.h>

/* -------------------------------------------------------------------------- */

#define NDIS_BUFFER_SIZE   2048

#define NDIS_TX_RATE_UNIT     1000000
#define NDIS_MEDIA_MAX_SPEED  (100 * NDIS_TX_RATE_UNIT)

/* -------------------------------------------------------------------------- */

#define DRIVER_STATE_INIT          0x00000001
#define DRIVER_STATE_RESET         0x00000002
#define DRIVER_STATE_PAUSE         0x00000004
#define DRIVER_STATE_PAUSED        0x00000008
#define DRIVER_STATE_RESTART       0x00000010
#define DRIVER_STATE_HANGED        0x80000000


#define TASK_OFFLOAD_CHECKSUM      0x00000001
#define TASK_OFFLOAD_IP_CHECKSUM   0x00000002
#define TASK_OFFLOAD_TCP_CHECKSUM  0x00000004
#define TASK_OFFLOAD_UDP_CHECKSUM  0x00000008


// Data structure for Wake Up Packet Pattern
#define MAX_PATTERN_LENGTH          128
#define MAX_WAKE_UP_PACKET          4

typedef struct tWAKE_UP_PACKET
{
    NDIS_PM_PACKET_PATTERN NdisPacketPattern;
    UCHAR MaskAndPattern[MAX_PATTERN_LENGTH + MAX_PATTERN_LENGTH / 8];
} WAKE_UP_PACKET, *PWAKE_UP_PACKET;


typedef struct _TCB
{
    SLIST_ENTRY          listEntry;
    PNET_BUFFER          NetBuffer;
    ULONG                BufferCount;
    ULONG                PacketLength;
    PVOID                pAdapter;
    PUCHAR               ScatterGatherBuffer;
    PSCATTER_GATHER_LIST pSGL;
} TCB, *PTCB;


typedef struct
{
    UINT                  ulAllocSize;
    UINT                  ulPhysical;
    BOOLEAN               bCached;
    PUCHAR                pAllocVirtual;
    PUCHAR                pVirtual;
    NDIS_PHYSICAL_ADDRESS AllocPhysicalAddr;
} SHARED_MEM, *PSHARED_MEM;


typedef struct _DMA_BUFFER
{
    SLIST_ENTRY           listEntry;
    SHARED_MEM            sm;
    UINT                  ulSize;
    NDIS_PHYSICAL_ADDRESS PhysicalAddr;
    PMDL                  Mdl;
    PNET_BUFFER_LIST      Nbl;
    PTCB                  Tcb;
    PNET_BUFFER           pBuffer;
} DMA_BUFFER, *PDMA_BUFFER;

#define DMA_BUFFER(pDesc)  ((PDMA_BUFFER)((pDesc)->pReserved))


// Get a pointer to a LIST_ENTRY for the send wait list, from an NB pointer
#define SEND_WAIT_LIST_FROM_NB(_NB) ((PLIST_ENTRY)&(_NB)->MiniportReserved[0])

// Get a pointer to a NB, from a pointer to a LIST_ENTRY on the send wait list
#define NB_FROM_SEND_WAIT_LIST(_ENTRY) (CONTAINING_RECORD(_ENTRY, NET_BUFFER, MiniportReserved[0]))

#define NBL_FROM_SEND_NB(_NB) (*((PNET_BUFFER_LIST*)&(_NB)->MiniportReserved[2]))
#define FRAME_TYPE_FROM_SEND_NB(_NB) (*((PULONG)&(_NB)->MiniportReserved[3]))


// Gets the number of outstanding TCBs/NET_BUFFERs associated with
// this NET_BUFFER_LIST
#define SEND_REF_FROM_NBL(_NBL) (*((PLONG)&(_NBL)->MiniportReserved[1]))

#define DESC_FROM_RCV_NBL(_NBL) ((_NBL)->MiniportReserved[0])
#define DMA_FROM_RCV_NBL(_NBL) (*((PDMA_BUFFER*)&(_NBL)->MiniportReserved[1]))


typedef struct
{
    NDIS_SPIN_LOCK listLock;
    SLIST_HEADER   listHead;

    /* Memory pool for buffers. */
    SHARED_MEM     MemPool;

    /* Array of buffers. */
    PDMA_BUFFER    BufArray;

    /* Current buffer being manipulated .*/
    PDMA_BUFFER    pCurrent;

    /* Number of buffers allocated. */
    int            cnAlloc;
} BUFFER_INFO, *PBUFFER_INFO;


typedef struct
{
    NDIS_SPIN_LOCK listLock;
    SLIST_HEADER   listHead;

    /* Array of buffers. */
    PTCB           TcbArray;
    NDIS_HANDLE    hDma;
    PUCHAR         ScatterGatherBuffer;
    ULONG          ScatterGatherSize;

    /* Number of buffers allocated. */
    int            cnAlloc;
} TCB_INFO, *PTCB_INFO;

/* -------------------------------------------------------------------------- */

typedef struct _NDIS_ADAPTER
{
    NDIS_HANDLE             hAdapter;

    NDIS_HANDLE             hInterrupt;

    NDIS_INTERFACE_TYPE     BusType;

    NDIS_SPIN_LOCK          lockAdapter;
    NDIS_SPIN_LOCK          lockReceive;
    NDIS_SPIN_LOCK          lockHardware;
    NDIS_EVENT              eventHardware;

    NDIS_HANDLE             hPeriodTimer;
    NDIS_HANDLE             hResetTimer;

    HARDWARE                Hardware;

    ULONG                   ulMacOptions;
    ULONG                   ulStatistics;
    ULONG                   ulPacketFilter;
    ULONG                   ulPacketsNotAllowed;
    ULONG                   ulMaxLookAhead;
    ULONG                   ulNdisMediaState;
    UINT                    OidsSize;

    // hardware resources
#ifdef _WIN64
    LONGLONG                MemoryAddress;
#else
    ULONG                   MemoryAddress;
#endif
    UINT                    ulMemorySize;
    UINT                    ulInterruptNumber;
    UINT                    ulInterruptVector;
    UINT                    ulIOBaseAddress;
    UINT                    ulIOLength;

#if defined(NDIS620_MINIPORT)
    NDIS_PM_CAPABILITIES    PmCapabilities;
#else
    NDIS_PNP_CAPABILITIES   PnpCapabilities;
#endif
    NDIS_DEVICE_POWER_STATE CurrentPowerState;
    NDIS_DEVICE_POWER_STATE NextPowerState;
    ULONG                   ulWakeUpEnable;
    ULONG                   ulWakeUpOk;
    ULONG                   ulWakeUpError;
    WAKE_UP_PACKET          WakeUpPacket[MAX_WAKE_UP_PACKET];
    int                     cnWakeUpPacket;

    ULONG                   ulTxChecksum;
    ULONG                   ulRxChecksum;
    ULONG                   ulUdpTestLimit;

    // member variables used for transmitting

    NDIS_HANDLE hNetBufListPool;

    SHARED_MEM              DescPool;

    BUFFER_INFO             RxBufInfo;
    BUFFER_INFO             TxBufInfo;
    TCB_INFO                TcbInfo;

    int                     nPME_wait;

    // List of net buffers to send that are waiting for a free TCB
    LIST_ENTRY              SendWaitList;
    NDIS_SPIN_LOCK          SendWaitListLock;

    // Number of transmit NBLs from the protocol that we still have
    volatile LONG           nBusySend;

    // Simple mutex to ensure only one CPU is sending at a time
    volatile LONG           TransmitAcquire;

    // Simplme mutex to ensure only one CPU is operating on interrupts
    volatile LONG           InterruptAcquire;
#ifdef DEBUG_INTR_LOCK
    volatile LONG           OperateAcquire;
    volatile int            InterruptOperate;
#endif

    UINT                    ulDriverState;
    ULONG                   PauseReason;

    LONG                    nNotSend;
    LONG                    nHangCnt;

    UINT                    nCopySend;

    BOOLEAN                 bCheckCounter;
    BOOLEAN                 bPacketFilterSet;
    BOOLEAN                 fSend;
} NDIS_ADAPTER;

typedef struct _NDIS_ADAPTER* PNDIS_ADAPTER;

#include <poppack.h>

/* -------------------------------------------------------------------------- */

PVOID AllocMemory (
    IN  NDIS_HANDLE hAdapter,
    IN  ULONG       Size );

VOID DriverSleep (
    IN  PNDIS_ADAPTER pAdapter );

VOID DriverWake (
    IN  PNDIS_ADAPTER           pAdapter,
    IN  NDIS_DEVICE_POWER_STATE LastPowerState );

VOID FreeMemory (
        PNDIS_ADAPTER pAdapter );

NDIS_STATUS InitAdapter (
        PNDIS_ADAPTER pAdapter,
    IN  PNDIS_MINIPORT_INIT_PARAMETERS InitParameters );

NDIS_STATUS SetupAdapter (
        PNDIS_ADAPTER pAdapter );

void ShutdownAdapter (
    IN  PNDIS_ADAPTER pAdapter );

void StartAdapter (
    IN  PNDIS_ADAPTER pAdapter );

BOOLEAN AdapterReset (
        PNDIS_ADAPTER pAdapter );

VOID DeRegisterAdapter (
    PNDIS_ADAPTER pAdapter );

NDIS_STATUS RegisterAdapter (
    PNDIS_ADAPTER pAdapter );

VOID IndicateLinkStatus (
    IN  PNDIS_ADAPTER pAdapter );

VOID SetTimer (
    IN  NDIS_HANDLE hTimer,
    IN  LONG        msec,
    IN  BOOLEAN     repeat );

BOOLEAN AcquireAdapter (
    IN  PNDIS_ADAPTER pAdapter,
    IN  BOOLEAN       fWait );

VOID ReleaseAdapter (
    IN  PNDIS_ADAPTER pAdapter );

VOID AdapterGetDescription (
    PVOID* ppSource,
    PUINT  puiLength );

ULONGLONG AdapterReadCounter (
    PNDIS_ADAPTER pAdapter,
    ULONG         OidCounter );

#define INTR_ACQ_ISR       0x00000001
#define INTR_ACQ_HANDLE    0x00000002
#define INTR_ACQ_BLOCK     0x00000004
#define INTR_ACQ_RESTORE   0x00000008
#define INTR_ACQ_CHECK     0x00000080

VOID AcquireInterrupt (
    PNDIS_ADAPTER pAdapter,
    int           operate );

VOID ReleaseInterrupt (
    PNDIS_ADAPTER pAdapter,
    int           operate );


void PrintMacAddress (
    PUCHAR bAddr );

NDIS_STATUS GetDeviceMode (
    IN  PNDIS_ADAPTER pAdapter,
    IN  NDIS_OID      Oid,
    IN  PVOID         pBuffer,
    IN  ULONG         ulLen,
    OUT PULONG        pulBytesWritten,
    OUT PULONG        pulBytesNeeded );

NDIS_STATUS SetDeviceMode (
    IN  PNDIS_ADAPTER pAdapter,
    IN  NDIS_OID      Oid,
    IN  PVOID         pBuffer,
    IN  ULONG         ulLen,
    OUT PULONG        pulBytesRead,
    OUT PULONG        pulBytesNeeded );


#ifndef INLINE
PDMA_BUFFER Alloc_DMA_Buffer (
    PBUFFER_INFO pBufInfo,
    PTDesc       pCurrent );

VOID Free_DMA_Buffer (
    PBUFFER_INFO pBufInfo,
    PDMA_BUFFER  pDma );

#else
__inline
PDMA_BUFFER Alloc_DMA_Buffer (
    PBUFFER_INFO pBufInfo,
    PTDesc       pCurrent )
{
    PDMA_BUFFER pDma;
    SLIST_ENTRY *pEntry;

    pEntry = NdisInterlockedPopEntrySList(&pBufInfo->listHead,
        &pBufInfo->listLock);
    ASSERT(pEntry != NULL);
    pDma = CONTAINING_RECORD(pEntry, DMA_BUFFER, listEntry);

    /* Link to current descriptor. */
    pCurrent->pReserved = pDma;
    return pDma;
}  // Alloc_DMA_Buffer


__inline
VOID Free_DMA_Buffer (
    PBUFFER_INFO pBufInfo,
    PDMA_BUFFER  pDma )
{
    NdisInterlockedPushEntrySList(&pBufInfo->listHead, &pDma->listEntry,
        &pBufInfo->listLock);
}
#endif


VOID TransmitDone (
    PNDIS_ADAPTER pAdapter );

VOID TransmitReset (
    PNDIS_ADAPTER pAdapter );

#endif
