/* ---------------------------------------------------------------------------
    Copyright (c) 2003-2013 Micrel, Inc.
    Copyright (c) 2015 Microchip Technology Inc.
   ---------------------------------------------------------------------------

    NdisDriver.c - NDIS driver functions.

    Author      Date        Description
    THa         01/26/11    Ported from NDIS 5.0 driver with code from
                            Microsoft sample driver.
    ---------------------------------------------------------------------------
*/


#include "target.h"
#include "hardware.h"
#include "NdisDevice.h"


DRIVER_INITIALIZE DriverEntry;

#pragma NDIS_INIT_FUNCTION(DriverEntry)
#pragma NDIS_PAGEABLE_FUNCTION(MiniportDriverUnload)
#pragma NDIS_PAGEABLE_FUNCTION(MiniportInitializeEx)
#pragma NDIS_PAGEABLE_FUNCTION(MiniportDevicePnPEventNotify)
#pragma NDIS_PAGEABLE_FUNCTION(MiniportHaltEx)

/* -------------------------------------------------------------------------- */

#if defined(DEF_KS8842)
#define DRV_NAME     "KSZ8842"
#elif defined(DEF_KS8841)
#define DRV_NAME     "KSZ8841"
#else
#error Either DEF_KS8841 or DEF_KS8842 has to be defined.
#endif
#define BUS_NAME     "PCI"
#define DRV_VERSION  "1.0.22"
#define DRV_RELDATE  "Nov 11, 2015"

#ifdef DBG
static char version[] =
    "Micrel " DRV_NAME " " BUS_NAME " " DRV_VERSION " (" DRV_RELDATE ")";
#endif

#if 1
INT MPDebugLevel = MP_LOUD;
#else
INT MPDebugLevel = MP_WARNING;
#endif

/* -------------------------------------------------------------------------- */

VOID IndicateLinkStatus (
    IN  PNDIS_ADAPTER pAdapter )
{
    NDIS_LINK_STATE LinkState;
    NDIS_STATUS_INDICATION Indication;
    PHARDWARE pHardware = &pAdapter->Hardware;

    NdisZeroMemory(&LinkState, sizeof(LinkState));
    NdisZeroMemory(&Indication, sizeof(Indication));

    LinkState.Header.Type = NDIS_OBJECT_TYPE_DEFAULT;
    LinkState.Header.Revision = NDIS_LINK_STATE_REVISION_1;
    LinkState.Header.Size = sizeof(NDIS_LINK_STATE);

    if (MediaStateConnected == pAdapter->ulNdisMediaState)
        LinkState.MediaConnectState = MediaConnectStateConnected;
    else
        LinkState.MediaConnectState = MediaConnectStateDisconnected;
    if (MediaConnectStateConnected == LinkState.MediaConnectState)
    {
        if (2 == pHardware->m_ulDuplex)
            LinkState.MediaDuplexState = MediaDuplexStateFull;
        else if (1 == pHardware->m_ulDuplex)
            LinkState.MediaDuplexState = MediaDuplexStateHalf;
        else
            LinkState.MediaDuplexState = MediaDuplexStateUnknown;
        LinkState.XmitLinkSpeed = LinkState.RcvLinkSpeed =
            pHardware->m_ulTransmitRate *
            (NDIS_TX_RATE_UNIT / HW_TX_RATE_UNIT);
        pAdapter->bCheckCounter = FALSE;
    }
    else
    {
        LinkState.MediaDuplexState = MediaDuplexStateUnknown;
        LinkState.XmitLinkSpeed = LinkState.RcvLinkSpeed =
            NDIS_LINK_SPEED_UNKNOWN;
        pAdapter->bCheckCounter = TRUE;
    }

    Indication.Header.Type = NDIS_OBJECT_TYPE_STATUS_INDICATION;
    Indication.Header.Revision = NDIS_STATUS_INDICATION_REVISION_1;
    Indication.Header.Size = NDIS_SIZEOF_STATUS_INDICATION_REVISION_1;

    Indication.SourceHandle = pAdapter->hAdapter;
    Indication.PortNumber = 0;
    Indication.StatusCode = NDIS_STATUS_LINK_STATE;
    Indication.StatusBuffer = &LinkState;
    Indication.StatusBufferSize = sizeof(LinkState);
    NdisMIndicateStatusEx(pAdapter->hAdapter, &Indication);

    if ((pAdapter->ulDriverState & DRIVER_STATE_RESET) &&
            MediaStateConnected == pAdapter->ulNdisMediaState)
    {
        pAdapter->ulDriverState &= ~DRIVER_STATE_RESET;
        NdisMResetComplete(pAdapter->hAdapter, NDIS_STATUS_SUCCESS, FALSE);
    }
}  // IndicateLinkStatus


static
NDIS_STATUS AllocTimer (
    IN  PNDIS_ADAPTER        pAdapter,
    IN  PNDIS_TIMER_FUNCTION Function,
    OUT PNDIS_HANDLE         pHandle )
{
    NDIS_STATUS Status;
    NDIS_TIMER_CHARACTERISTICS Timer;

    NdisZeroMemory(&Timer, sizeof(Timer));

    Timer.Header.Type = NDIS_OBJECT_TYPE_TIMER_CHARACTERISTICS;
    Timer.Header.Size = NDIS_SIZEOF_TIMER_CHARACTERISTICS_REVISION_1;
    Timer.Header.Revision = NDIS_TIMER_CHARACTERISTICS_REVISION_1;

    Timer.AllocationTag = MEM_TAG;
    Timer.TimerFunction = Function;
    Timer.FunctionContext = pAdapter;

    Status = NdisAllocateTimerObject(pAdapter->hAdapter, &Timer, pHandle);
    return Status;
}  // AllocTimer


VOID SetTimer (
    IN  NDIS_HANDLE hTimer,
    IN  LONG        msec,
    IN  BOOLEAN     repeat )
{
    LARGE_INTEGER liRetryTime;
    LONG RepeatTime = 0;

    // in 100ns increments
    liRetryTime.QuadPart = -1000LL * msec;
    if (repeat)
        RepeatTime = msec;
    NdisSetTimerObject(hTimer, liRetryTime, RepeatTime, NULL);
}  // SetTimer


static
VOID PeriodTimerDpc (
    IN  PVOID SystemSpecific1,
    IN  PVOID hAdapterContext,
    IN  PVOID SystemSpecific2,
    IN  PVOID SystemSpecific3 )
{
    // get the pointer to the adapter from the context handle
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hAdapterContext;
    PHARDWARE pHardware = &pAdapter->Hardware;

    if (!AcquireAdapter(pAdapter, FALSE))
    {
        return;
    }

    if (!pHardware->m_bLinkIntWorking)
        SwitchGetLinkStatus(pHardware);

    if (PortReadCounters(pHardware, MAIN_PORT))
        goto monitor_release;

#ifdef DEF_KS8842
    if (PortReadCounters(pHardware, OTHER_PORT))
        goto monitor_release;
    if (PortReadCounters(pHardware, HOST_PORT))
        goto monitor_release;
#endif

    if (pAdapter->nPME_wait)
    {
        if (++pAdapter->nPME_wait >= 2) {
            HardwareClearWolPMEStatus(pHardware);

#ifdef DBG
            DbgPrint("clear PME."NEWLINE);
#endif
            pAdapter->nPME_wait = 0;
        }
    }
    else if (HardwareCheckWolPMEStatus(pHardware))
    {
        pAdapter->nPME_wait = 1;

#ifdef DBG
        DbgPrint("PME asserted."NEWLINE);
#endif
    }

monitor_release:

    if (pAdapter->fSend || pAdapter->nNotSend)
    {
        TransmitDone(pAdapter);
    }

    ReleaseAdapter(pAdapter);
    pAdapter->bCheckCounter = FALSE;

    if (pAdapter->ulNdisMediaState != pHardware->m_ulHardwareState)
    {
        pAdapter->ulNdisMediaState = pHardware->m_ulHardwareState;
        IndicateLinkStatus(pAdapter);
    }
}  // PeriodTimerDpc


static
VOID ResetCompleteTimerDpc (
    IN  PVOID SystemSpecific1,
    IN  PVOID hAdapterContext,
    IN  PVOID SystemSpecific2,
    IN  PVOID SystemSpecific3 )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hAdapterContext;

    if ((pAdapter->ulDriverState & DRIVER_STATE_PAUSE))
    {
        pAdapter->ulDriverState &= ~DRIVER_STATE_PAUSE;
        pAdapter->ulDriverState |= DRIVER_STATE_PAUSED;
        NdisMPauseComplete(pAdapter->hAdapter);
    }
    if ((pAdapter->ulDriverState & DRIVER_STATE_RESTART))
    {
        pAdapter->ulDriverState &= ~DRIVER_STATE_RESTART;
        NdisMRestartComplete(pAdapter->hAdapter, NDIS_STATUS_SUCCESS);
    }
    if ((pAdapter->ulDriverState & DRIVER_STATE_RESET))
    {
        pAdapter->ulDriverState &= ~DRIVER_STATE_RESET;
        NdisMResetComplete(pAdapter->hAdapter, NDIS_STATUS_SUCCESS, FALSE);
    }
    if ((pAdapter->ulDriverState & DRIVER_STATE_INIT))
    {
        pAdapter->ulDriverState &= ~DRIVER_STATE_INIT;
        SetTimer(pAdapter->hPeriodTimer, 500, TRUE);
    }
}  // ResetCompleteTimerDpc

/* -------------------------------------------------------------------------- */

static struct hw_fn* ks8842_fn = NULL;

/*
    DriverSleep

Description:
    This routine halts the adapter when the driver is put to sleep.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

Return (None):
*/

VOID DriverSleep (
    IN  PNDIS_ADAPTER pAdapter )
{
    // cancel the period timer
    NdisCancelTimerObject(pAdapter->hPeriodTimer);
    NdisCancelTimerObject(pAdapter->hResetTimer);

    ShutdownAdapter(pAdapter);
}  // DriverSleep


/*
    DriverWake

Description:
    This routine starts the adapter when the driver is woke up.

Parameters:
    PNDIS_ADAPTER pAdapter
        Pointer to adapter information structure.

    NDIS_DEVICE_POWER_STATE LastPowerState
        The last power state used.  Device is reset when resumed from suspend.

Return (None):
*/

VOID DriverWake (
    IN  PNDIS_ADAPTER           pAdapter,
    IN  NDIS_DEVICE_POWER_STATE LastPowerState )
{
    if (NdisDeviceStateD3 == LastPowerState)
        AdapterReset(pAdapter);
    else
        StartAdapter(pAdapter);

    pAdapter->ulDriverState |= DRIVER_STATE_INIT;
    SetTimer(pAdapter->hResetTimer, 1000, FALSE);
}  // DriverWake


/*
    MiniportHaltEx

Description:
    This routine halts the adapter when the driver is stopped.

Parameters:
    NDIS_HANDLE hContext
        Handle to adapter context containing adapter information.

    NDIS_HALT_ACTION action
        The halt action specified.

Return (None):
*/

static
VOID MiniportHaltEx (
    IN  NDIS_HANDLE      hContext,
    IN  NDIS_HALT_ACTION action )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hContext;
#ifdef DEBUG_MINIPORT_HALT
    PCHAR reason;
#endif

    PAGED_CODE();

#ifdef DEBUG_MINIPORT_HALT
    DbgPrint("MiniportHalt"NEWLINE);
    switch (action) {
        #undef MAKECASE
        #define MAKECASE(halt) case halt: reason = #halt "\n"; break;

        MAKECASE(NdisHaltDeviceDisabled)
        MAKECASE(NdisHaltDeviceInstanceDeInitialized)
        MAKECASE(NdisHaltDevicePoweredDown)
        MAKECASE(NdisHaltDeviceSurpriseRemoved)
        MAKECASE(NdisHaltDeviceFailed)
        MAKECASE(NdisHaltDeviceInitializationFailed)
        MAKECASE(NdisHaltDeviceStopped)
    }
    if (reason)
    {
        DEBUGP(MP_LOUD, ("%s", reason));
    }
#endif

    DriverSleep(pAdapter);
    HardwareSetPowerSaving(&pAdapter->Hardware, TRUE);

    // Disconnect the interrupt line
    NdisMDeregisterInterruptEx(pAdapter->hInterrupt);

    // Pause, waiting for any DPC stuff to clear.
    NdisMSleep(250000);

    DeRegisterAdapter(pAdapter);

    NdisFreeSpinLock(&pAdapter->lockAdapter);
    NdisFreeSpinLock(&pAdapter->lockReceive);
    NdisFreeSpinLock(&pAdapter->lockHardware);
    NdisFreeTimerObject(pAdapter->hPeriodTimer);
    NdisFreeTimerObject(pAdapter->hResetTimer);
    FreeMemory(pAdapter);
    NdisFreeMemory(pAdapter, sizeof(NDIS_ADAPTER), 0);
}  // MiniportHaltEx


/*
    MiniportShutdownEx

Description:
    This routine stops the adapter when the driver is stopped.

    Notes: MiniportShutdownEx SHOULD CALL NO NDISXXX FUNCTIONS.

Parameters:
    NDIS_HANDLE hContext
        Handle to adapter context containing adapter information.

    NDIS_SHUTDOWN_ACTION action
        The shutdown action specified.

Return (None):
*/

static
VOID MiniportShutdownEx (
    IN  NDIS_HANDLE          hContext,
    IN  NDIS_SHUTDOWN_ACTION action )
{
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hContext;

#if DBG
    DbgPrint("MiniportShutdown"NEWLINE);
    if (NdisShutdownPowerOff == action)
        DbgPrint("power off"NEWLINE);
    else if (NdisShutdownBugCheck == action)
        DbgPrint("bug check"NEWLINE);
#endif

    ShutdownAdapter(pAdapter);
    HardwareSetPowerSaving(&pAdapter->Hardware, TRUE);
}  // MiniportShutdownEx


/*
    MiniportDevicePnPEventNotify

Description:
    This routine notifies the device about PnP events.

Parameters:
    NDIS_HANDLE hContext
        Handle to adapter context containing adapter information.

    NDIS_DEVICE_PNP_EVENT PnPEvent
        PnP event information.

Return (None):
*/

static
VOID MiniportDevicePnPEventNotify (
    IN  NDIS_HANDLE           hContext,
    IN  PNET_DEVICE_PNP_EVENT PnPEvent )
{
#if 0
    PNDIS_ADAPTER pAdapter = (PNDIS_ADAPTER) hContext;
#endif

    PAGED_CODE();

    switch (PnPEvent->DevicePnPEvent)
    {
        case NdisDevicePnPEventSurpriseRemoved:
            //
            // Called when NDIS receives IRP_MN_SUPRISE_REMOVAL.
            // NDIS calls MiniportHalt function after this call returns.
            //
#if DBG
            DBG_PRINT("Removed!"NEWLINE);
#endif
            break;

        case NdisDevicePnPEventPowerProfileChanged:
            //
            // After initializing a miniport driver and after miniport driver
            // receives an OID_PNP_SET_POWER notification that specifies
            // a device power state of NdisDeviceStateD0 (the powered-on state),
            // NDIS calls the miniport's MiniportPnPEventNotify function with
            // PnPEvent set to NdisDevicePnPEventPowerProfileChanged.
            //
            if (PnPEvent->InformationBufferLength == sizeof(ULONG))
            {
                ULONG profile = *((PULONG) PnPEvent->InformationBuffer);

                if (NdisPowerProfileBattery == profile)
                {
#if DBG
                    DBG_PRINT("battery"NEWLINE);
#endif
                }
                else if (NdisPowerProfileAcOnLine == profile)
                {
#if DBG
                    DBG_PRINT("AC"NEWLINE);
#endif
                }
            }
            break;
    }
}  // MiniportDevicePnPEventNotify


static
NDIS_STATUS RegisterInterrupt (
    IN  PNDIS_ADAPTER pAdapter )
{
    NDIS_MINIPORT_INTERRUPT_CHARACTERISTICS Interrupt;
    NDIS_STATUS Status;

    NdisZeroMemory(&Interrupt,
        sizeof(NDIS_MINIPORT_INTERRUPT_CHARACTERISTICS));
    Interrupt.Header.Type = NDIS_OBJECT_TYPE_MINIPORT_INTERRUPT;
    Interrupt.Header.Revision = NDIS_MINIPORT_INTERRUPT_REVISION_1;
    Interrupt.Header.Size =
        NDIS_SIZEOF_MINIPORT_INTERRUPT_CHARACTERISTICS_REVISION_1;

    Interrupt.DisableInterruptHandler = MiniportDisableInterruptEx;
    Interrupt.EnableInterruptHandler = MiniportEnableInterruptEx;
    Interrupt.InterruptHandler = MiniportISR;
    Interrupt.InterruptDpcHandler = MiniportHandleInterrupt;
    Status = NdisMRegisterInterruptEx(pAdapter->hAdapter, pAdapter, &Interrupt,
        &pAdapter->hInterrupt);
    return Status;
}  // RegisterInterrupt


/*
    MiniportInitializeEx

Description:
    This function initializes the adapter and registers resources with the
    NDIS wrapper.

    The MiniportInitializeEx function is a required function that sets up a
    NIC (or virtual NIC) for network I/O operations, claims all hardware
    resources necessary to the NIC in the registry, and allocates resources
    the driver needs to carry out network I/O operations.

    MiniportInitializeEx runs at IRQL = PASSIVE_LEVEL.

Parameters:
    NDIS_HANDLE hAdapter
        Handle to adapter to pass to the NDIS wrapper.

    NDIS_HANDLE hContext
        Handle of context information passed to the device, normally not used.

    PNDIS_MINIPORT_INIT_PARAMETERS InitParameters
        Device init parameters passed to the driver.

Return (NDIS_STATUS):
    NDIS_STATUS_SUCCESS if successful; otherwise an error code indicating
    failure.
*/

static
NDIS_STATUS MiniportInitializeEx (
    IN  NDIS_HANDLE                    hAdapter,
    IN  NDIS_HANDLE                    hContext,
    IN  PNDIS_MINIPORT_INIT_PARAMETERS InitParameters )
{
    PNDIS_ADAPTER pAdapter;
    NDIS_STATUS Status;
    UCHAR bInterrupt = FALSE;

    UNREFERENCED_PARAMETER(hContext);
    PAGED_CODE();

    // Allocate memory for the adapter block now.

    pAdapter = AllocMemory(hAdapter, sizeof(NDIS_ADAPTER));
    if (!pAdapter)
    {
        return NDIS_STATUS_FAILURE;
    }

    // Clear out the adapter block, which sets all default values to FALSE
    // or NULL.

    NdisZeroMemory(pAdapter, sizeof(NDIS_ADAPTER));

    // Save the adapter handle furnished by NDIS.
    pAdapter->hAdapter = hAdapter;

    pAdapter->Hardware.m_hwfn = ks8842_fn;

    // Allocate the spin locks.
    NdisAllocateSpinLock(&pAdapter->lockAdapter);
    NdisAllocateSpinLock(&pAdapter->lockReceive);
    NdisAllocateSpinLock(&pAdapter->lockHardware);
    NdisInitializeEvent(&pAdapter->eventHardware);

    // Process the configuration parameters.
    Status = InitAdapter(pAdapter, InitParameters);
    if (Status != NDIS_STATUS_SUCCESS)
    {
        goto MiniportInitializeMemoryError;
    }

    // Register the adapter.
    Status = RegisterAdapter(pAdapter);
    if (Status != NDIS_STATUS_SUCCESS)
    {
        goto MiniportInitializeError;
    }

    Status = RegisterInterrupt(pAdapter);
    if (Status != NDIS_STATUS_SUCCESS)
    {
        NdisWriteErrorLogEntry(pAdapter->hAdapter,
            NDIS_ERROR_CODE_INTERRUPT_CONNECT, 0);
        goto MiniportInitializeError;
    }
    bInterrupt = TRUE;

    Status = SetupAdapter(pAdapter);
    if (Status != NDIS_STATUS_SUCCESS)
    {
        goto MiniportInitializeError;
    }
    Status = AllocTimer(pAdapter, ResetCompleteTimerDpc,
        &pAdapter->hResetTimer);
    Status = AllocTimer(pAdapter, PeriodTimerDpc, &pAdapter->hPeriodTimer);

    DriverWake(pAdapter, NdisDeviceStateD0);

    return NDIS_STATUS_SUCCESS;

MiniportInitializeError:
    if (bInterrupt)
        NdisMDeregisterInterruptEx(pAdapter->hInterrupt);
    DeRegisterAdapter(pAdapter);

MiniportInitializeMemoryError:
    FreeMemory(pAdapter);

    NdisFreeSpinLock(&pAdapter->lockAdapter);
    NdisFreeSpinLock(&pAdapter->lockReceive);
    NdisFreeSpinLock(&pAdapter->lockHardware);
    NdisFreeMemory(pAdapter, sizeof(NDIS_ADAPTER), 0);
    return Status;
}  // MiniportInitializeEx

/* -------------------------------------------------------------------------- */

static NDIS_HANDLE NdisMiniportHandle = NULL;

static
VOID MiniportDriverUnload (
    IN  PDRIVER_OBJECT DriverObject )
{
#if DBG
    DBG_PRINT("MiniportDriverUnload"NEWLINE);
#endif
    PAGED_CODE();

    NdisMDeregisterMiniportDriver(NdisMiniportHandle);
    if (ks8842_fn)
    {
        NdisFreeMemory(ks8842_fn, sizeof(struct hw_fn), 0);
        ks8842_fn = NULL;
    }
}  /* MiniportDriverUnload */


/*
    DriverEntry

Description:
    This function is the main entry point of the driver.

Parameters:
    PDRIVER_OBJECT pDriverObject
        Pointer to driver object.

    PUNICODE_STRING RegistryPath
        Pointer to unicode string containing the register path that stores
        the configuration of the driver.

Return (NTSTATUS):
    STATUS_SUCCESS if successful; otherwise STATUS_UNSUCCESSFUL.
*/

#ifdef __cplusplus
extern "C"
#endif
NTSTATUS DriverEntry (
    IN  PDRIVER_OBJECT  DriverObject,
    IN  PUNICODE_STRING RegistryPath )
{
    // Characteristics table for this driver.
    NDIS_MINIPORT_DRIVER_CHARACTERISTICS Driver;
    NDIS_STATUS Status;

#ifdef DBG
    DBG_PRINT("%s"NEWLINE, version);
#endif
    PAGED_CODE();

    // Clear the characteristics table to ensure there's no bogus data passed
    // to NDIS.
    NdisZeroMemory(&Driver, sizeof(Driver));

    // Initialize the Miniport driver characteristics for the call to
    // NdisMRegisterMiniportDriver.
    Driver.Header.Type = NDIS_OBJECT_TYPE_MINIPORT_DRIVER_CHARACTERISTICS,
    Driver.Header.Size = sizeof(NDIS_MINIPORT_DRIVER_CHARACTERISTICS);
#if defined(NDIS620_MINIPORT)
    Driver.Header.Revision = NDIS_MINIPORT_DRIVER_CHARACTERISTICS_REVISION_2;
#else
    Driver.Header.Revision = NDIS_MINIPORT_DRIVER_CHARACTERISTICS_REVISION_1;
#endif

    Driver.MajorNdisVersion             = NDIS_MINIPORT_MAJOR_VERSION;
    Driver.MinorNdisVersion             = NDIS_MINIPORT_MINOR_VERSION;
    Driver.MajorDriverVersion           = DRIVER_MAJOR_VERSION;
    Driver.MinorDriverVersion           = DRIVER_MINOR_VERSION;

#if 0
    Driver.SetOptionsHandler            = MiniportSetOptions;
#endif

    Driver.InitializeHandlerEx          = MiniportInitializeEx;
    Driver.HaltHandlerEx                = MiniportHaltEx;
    Driver.CheckForHangHandlerEx        = MiniportCheckForHangEx;
    Driver.ResetHandlerEx               = MiniportResetEx;

    Driver.UnloadHandler                = MiniportDriverUnload,

    Driver.PauseHandler                 = MiniportPause;
    Driver.RestartHandler               = MiniportRestart;
    Driver.OidRequestHandler            = MiniportOidRequest;
    Driver.CancelOidRequestHandler      = MiniportCancelOidRequest;

    Driver.ReturnNetBufferListsHandler  = MiniportReturnNetBufferLists;
    Driver.SendNetBufferListsHandler    = MiniportSendNetBufferLists;

    Driver.ShutdownHandlerEx            = MiniportShutdownEx;
    Driver.CancelSendHandler            = MiniportCancelSend;
    Driver.DevicePnPEventNotifyHandler  = MiniportDevicePnPEventNotify;

    Status = NdisMRegisterMiniportDriver(DriverObject, RegistryPath, NULL,
        &Driver, &NdisMiniportHandle);

    if (NDIS_STATUS_SUCCESS == Status)
    {
        ks8842_fn = AllocMemory(NdisMiniportHandle, sizeof(struct hw_fn));
        HardwareSetupFunc(ks8842_fn);
    }

#if DBG
    DbgPrint("%p = ", DriverObject);
    switch (Status)
    {
        case NDIS_STATUS_BAD_CHARACTERISTICS:
            DbgPrint("NDIS_STATUS_BAD_CHARACTERISTICS\n");
            break;
        case NDIS_STATUS_BAD_VERSION:
            DbgPrint("NDIS_STATUS_BAD_VERSION\n");
            break;
        case NDIS_STATUS_RESOURCES:
            DbgPrint("NDIS_STATUS_RESOURCES\n");
            break;
        case NDIS_STATUS_FAILURE:
            DbgPrint("NDIS_STATUS_FAILURE\n");
            break;
        case NDIS_STATUS_SUCCESS:
            DbgPrint("NDIS_STATUS_SUCCESS\n");
            break;
        default:
            DbgPrint("NDIS Unknown:%08x\n", Status);
            break;
    }
#endif

    return(Status == NDIS_STATUS_SUCCESS ? STATUS_SUCCESS :
        STATUS_UNSUCCESSFUL);
}  // DriverEntry
